@@ -4,20 +4,19 @@
  *
  * Copyright (C) 2009 - Jesse van den Kieboom
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * gtksourceview is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * gtksourceview is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor,
- * Boston, MA  02110-1301  USA
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #include "gtksourcegutter-private.h"
@@ -74,7 +73,7 @@ struct _GtkSourceGutterPrivate
 {
 	GtkSourceView *view;
 	GtkTextWindowType window_type;
-
+	gint size;
 	GList *renderers;
 	guint signals[LAST_EXTERNAL_SIGNAL];
 };
@@ -206,6 +205,8 @@ set_view (GtkSourceGutter *gutter,
 {
 	gutter->priv->view = view;
 
+	gutter->priv->size = -1;
+
 	g_object_weak_ref (G_OBJECT (view),
 	                   (GWeakNotify)view_notify,
 	                   gutter);
@@ -275,7 +276,7 @@ revalidate_size (GtkSourceGutter *gutter)
 		gtk_text_view_set_border_window_size (GTK_TEXT_VIEW (gutter->priv->view),
 		                                      gutter->priv->window_type,
 		                                      1);
-		do_redraw (gutter);
+		gutter->priv->size = -1;
 	}
 	else if (window && !gutter->priv->renderers)
 	{
@@ -376,7 +377,7 @@ gtk_source_gutter_class_init (GtkSourceGutterClass *klass)
 			      3,
 			      GTK_TYPE_CELL_RENDERER,
 			      GTK_TYPE_TEXT_ITER,
-			      G_TYPE_POINTER);
+			      GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE);
 
 	/**
 	 * GtkSourceGutter::query-tooltip:
@@ -409,6 +410,8 @@ static void
 gtk_source_gutter_init (GtkSourceGutter *self)
 {
 	self->priv = GTK_SOURCE_GUTTER_GET_PRIVATE (self);
+
+	self->priv->size = -1;
 }
 
 static gint
@@ -682,7 +685,6 @@ gtk_source_gutter_queue_draw (GtkSourceGutter *gutter)
 	revalidate_size (gutter);
 }
 
-/* Callbacks */
 static gint
 calculate_size (GtkSourceGutter  *gutter,
                 Renderer         *renderer)
@@ -844,12 +846,33 @@ on_view_expose_event (GtkSourceView   *view,
 	text_view = GTK_TEXT_VIEW (view);
 	sizes = g_array_new (FALSE, FALSE, sizeof (gint));
 
+	/* This is fairly ugly, but we could not find a better way to
+	 * do it: renderers could have changed size and they do not have
+	 * a way to signal it. So on expose we revalidate the size and
+	 * requeue another expose.
+	 * To see if the size has changed we test the size of only the
+	 * gutter itself since the full border window size is not under
+	 * our control (see e.g bug #589382).
+	 * This also means that if the user manually forces the
+	 * border_window_size to a value smaller than the gutter, things
+	 * will not be drawn properly. */
 	size = calculate_sizes (gutter, sizes);
-
-	if (gtk_text_view_get_border_window_size (text_view, gutter->priv->window_type) != size)
+	if (gutter->priv->size != size)
 	{
+
+		gint border_size;
+
+		border_size = gtk_text_view_get_border_window_size (text_view, gutter->priv->window_type);
+
+		if (gutter->priv->size < 0)
+			border_size += size;
+		else
+			border_size = MAX (0, border_size - gutter->priv->size) + size;
+
+		gutter->priv->size = size;
+
 		/* Will trigger a new expose */
-		gtk_text_view_set_border_window_size (text_view, gutter->priv->window_type, size);
+		gtk_text_view_set_border_window_size (text_view, gutter->priv->window_type, border_size);
 	}
 	else
 	{
